from typing import Any, Dict, Optional

import numpy as np
import pandas as pd
from single_sample_gsea import ss_gsea

from Orange.data import Table, Domain, ContinuousVariable
from Orange.widgets import gui
from Orange.widgets.widget import Msg, Input, Output, OWWidget
from Orange.widgets.settings import Setting
from Orange.widgets.utils.concurrent import TaskState, ConcurrentWidgetMixin
from Orange.widgets.utils.widgetpreview import WidgetPreview

from orangecontrib.bioinformatics.geneset import GeneSets
from orangecontrib.bioinformatics.widgets.utils.data import (
    TableAnnotation,
    check_table_annotation,
)


def ss_mean(df, gene_sets, callback=None):
    scores = {}

    for gs_name, gs_genes in gene_sets.items():
        genes = [gene for gene in gs_genes if gene in df.index]
        scores[gs_name] = df.loc[genes].mean()

        if callback is not None:
            callback()

    df_results = pd.DataFrame(scores)
    df_results.index = df.columns
    return df_results


scoring_methods = [('mean', ss_mean), ('ssGSEA', ss_gsea)]


def worker(data: Table, gene_sets, selected_method: int, state: TaskState):
    progress_steps = iter(np.linspace(0, 100, len(gene_sets)))

    def callback():
        state.set_progress_value(next(progress_steps))

    # Prepare data
    gene_id_attribute = data.attributes[TableAnnotation.gene_id_attribute]
    genes = [attr.attributes[gene_id_attribute] for attr in data.domain.attributes]
    data_x = np.transpose(data.X)
    df = pd.DataFrame(data_x)
    df.index = genes

    _, method = scoring_methods[selected_method]
    return method(df, gene_sets, callback=callback)


class OWSingleSampleScoring(OWWidget, ConcurrentWidgetMixin):
    name = 'Single sample scoring'
    description = 'Scoring gene sets by single sample.'
    icon = 'icons/OWSingleSampleScoring.svg'
    priority = 150
    keywords = ['single sample scoring', 'gene sets', 'ssGSEA']
    want_main_area = False

    scoring_method: int = Setting(1)
    auto_commit: bool = Setting(True, schema_only=True)

    class Inputs:
        data = Input('Data', Table)
        gene_sets = Input('Gene Sets', (Table, GeneSets))

    class Outputs:
        data = Output('Data', Table)

    class Warning(OWWidget.Warning):
        wrong_data_format = Msg(
            'Please transpose the data. The widget expects '
            'samples as rows and genes as columns.'
        )

    def __init__(self):
        OWWidget.__init__(self)
        ConcurrentWidgetMixin.__init__(self)

        self.data: Optional[Table] = None
        self.gene_sets: Optional[Dict[str, set]] = None

        box = gui.widgetBox(self.controlArea, 'Method', margin=0)
        gui.comboBox(
            box,
            self,
            'scoring_method',
            items=[label for label, _ in scoring_methods],
            callback=self.commit.deferred,
        )

        gui.rubber(self.controlArea)
        self.commit_button = gui.auto_commit(
            self.buttonsArea, self, 'auto_commit', '&Commit', box=False
        )

    @Inputs.data
    @check_table_annotation
    def set_data(self, data: Table):
        self.Warning.wrong_data_format.clear()

        if not data:
            return

        gene_as_attr_name = data.attributes[TableAnnotation.gene_as_attr_name]
        if not gene_as_attr_name:
            self.Warning.wrong_data_format()
            self.sample_column_box.hide()
            return

        self.data = data
        self.start_worker()

    @Inputs.gene_sets
    def set_gene_sets(self, gene_sets):
        self.gene_sets = None
        if not gene_sets:
            return
        self.gene_sets = {gs.name: gs.genes for gs in gene_sets}

        self.start_worker()

    @gui.deferred
    def commit(self):
        self.start_worker()

    def start_worker(self):
        if not self.data or not self.gene_sets:
            return

        self.start(worker, self.data, self.gene_sets, self.scoring_method)

    def on_done(self, worker_result: pd.DataFrame):
        table = None

        if worker_result is not None:
            scores = []
            for column in worker_result.columns:
                var = ContinuousVariable(column)
                var.attributes['Score'] = 'Score'
                scores.append(var)

            domain = Domain(
                self.data.domain.attributes,
                metas=self.data.domain.metas + tuple(scores),
                class_vars=self.data.domain.class_vars,
            )

            table = self.data.transform(domain)
            with table.unlocked():
                table[:, scores] = worker_result.values

        self.Outputs.data.send(table)

    def on_partial_result(self, result: Any) -> None:
        pass

    def on_exception(self, ex):
        raise ex


if __name__ == '__main__':
    previewer = WidgetPreview(OWSingleSampleScoring)
    previewer.run(Table('iris'))



"""
DBGET Database Interface
========================

"""
from __future__ import absolute_import

import re
import sys
from contextlib import closing

from orangecontrib.bioinformatics.kegg import api, entry
from orangecontrib.bioinformatics.kegg.entry import fields


def iter_take(source_iter, n):
    """
    Return a list of the first `n` items in `source_iter`.
    """
    source_iter = iter(source_iter)
    return [item for _, item in zip(range(n), source_iter)]


def batch_iter(source_iter, n):
    """
    Split the `source_iter` into batches of size `n`.
    """
    source_iter = iter(source_iter)
    while True:
        batch = iter_take(source_iter, n)
        if batch:
            yield batch
        else:
            break


def chain_iter(chains_iter):
    for iter in chains_iter:
        for element in iter:
            yield element


# TODO: DBDataBase should be able to be constructed from a flat text
# entry file. The precache etc. should be moved in caching api, that creates
# simple file system hierarchy where the flat database is saved (with db
# release string), e.g.
# genes/hsa.dbget
# genes/hsa.release
# genes/sce.dbget
# path.dbget
# module.dbget
# ligand/compound.dbget


class DBDataBase(object):
    """
    Base class for a DBGET database interface.

    """

    #: ENTRY_TYPE constructor (a :class:`~.entry.DBEntry` subclass). This
    #: should be redefined in subclasses.
    ENTRY_TYPE = entry.DBEntry

    #: A database name/abbreviation (e.g. 'pathway'). Needs to be set in a
    #: subclass or object instance's constructor before calling the base.
    #: __init__
    DB = None

    def __init__(self, **kwargs):
        if not self.DB:
            raise TypeError("Cannot make an instance of abstract base " "class %r." % type(self).__name__)

        self.api = api.CachedKeggApi()
        self._info = None
        # TODO invalidate cache by KEGG release
        # self.api.set_default_release(self.info.release)
        self._keys = []

    @property
    def info(self):  # lazy info loading
        if not self._info:
            self._info = self.api.info(self.DB)
        return self._info

    def iterkeys(self):
        """
        Return an iterator over the `keys`.
        """
        return iter(self._keys)

    def iteritems(self):
        """
        Return an iterator over the `items`.
        """
        batch_size = 100
        iterkeys = self.iterkeys()
        return chain_iter(zip(batch, self.batch_get(batch)) for batch in batch_iter(iterkeys, batch_size))

    def itervalues(self):
        """
        Return an iterator over all :obj:`DBDataBase.ENTRY_TYPE` instances.
        """
        batch_size = 100
        iterkeys = self.iterkeys()
        return chain_iter(self.batch_get(batch) for batch in batch_iter(iterkeys, batch_size))

    if sys.version_info < (3,):

        def keys(self):
            """
            Return a list of database keys. These are unique KEGG identifiers
            that can be used to query the database.
            """
            return list(self._keys)

        def values(self):
            """
            Return a list of all :obj:`DBDataBase.ENTRY_TYPE` instances.
            """
            return self.batch_get(self.keys())

        def items(self):
            """
            Return a list of all (key, :obj:`DBDataBase.ENTRY_TYPE` instance)
            tuples.
            """
            return list(zip(self.keys(), self.batch_get(self.keys())))

    else:

        def keys(self):
            """
            Return an iterator over all database keys. These are unique
            KEGG identifiers that can be used to query the database.
            """
            return iter(self._keys)

        def values(self):
            """
            Return an iterator over all :obj:`DBDataBase.ENTRY_TYPE` instances.
            """
            return self.itervalues()

        def items(self):
            """
            Return an iterator over all (key, :obj:`DBDataBase.ENTRY_TYPE`)
            tuples.
            """
            return self.iteritems()

    def get(self, key, default=None):
        """
        Return an :obj:`DBDataBase.ENTRY_TYPE` instance for the `key`.
        Raises :class:`KeyError` if not found.

        """
        try:
            return self.__getitem__(key)
        except KeyError:
            return default

    def has_key(self, key):
        return self.__contains__(key)

    def __getitem__(self, key):
        e = self.get_entry(key)
        if e is None:
            raise KeyError(key)
        else:
            return e

    def __contains__(self, key):
        return key in set(self.keys())

    def __len__(self):
        return len(self.keys())

    def __iter__(self):
        return iter(self.keys())

    def get_text(self, key):
        """
        Return the database entry for `key` as plain text.
        """
        key = self._add_db(key)
        return self.api.get([key])

    def get_entry(self, key):
        """
        Return the database entry for `key` as an instance of `ENTRY_TYPE`.
        """
        text = self.get_text(key)
        if not text or text == "None":
            return None
        else:
            return self.ENTRY_TYPE(text)

    def find(self, name):
        """
        Find `name` using kegg `find` api.
        """
        res = self.api.find(self.DB, name).splitlines()
        return [r.split(" ", 1)[0] for r in res]

    def pre_cache(self, keys=None, batch_size=10, progress_callback=None):
        """
        Retrieve all the entries for `keys` and cache them locally for faster
        subsequent retrieval. If `keys` is ``None`` then all entries will be
        retrieved.

        """
        if not isinstance(self.api, api.CachedKeggApi):
            raise TypeError("Not an instance of api.CachedKeggApi")

        if batch_size > 10 or batch_size < 1:
            raise ValueError("Invalid batch_size")

        if keys is None:
            keys = self.keys()

        keys = map(self._add_db, keys)

        get = self.api.get

        # drop all keys with a valid cache entry to minimize the number
        # of 'get' requests.
        with closing(get.cache_store()) as store:

            def is_uncached(key):
                cache_key = get.key_from_args((key,))
                return not get.key_has_valid_cache(cache_key, store)

            keys = [key for key in keys if is_uncached(key)]

        start = 0

        while start < len(keys):
            batch = keys[start : start + batch_size]
            self.api.get(batch)

            if progress_callback:
                progress_callback(100.0 * start / len(keys))

            start += batch_size

    def batch_get(self, keys):
        """
        Batch retrieve all entries for keys. This can be significantly
        faster then getting each entry separately especially if entries
        are not yet cached.

        """
        entries = []
        batch_size = 10
        keys = list(map(self._add_db, keys))

        # Precache the entries first
        self.pre_cache(keys)

        start = 0
        while start < len(keys):
            batch = keys[start : start + batch_size]
            batch_entries = self.api.get(batch)
            if batch_entries is not None:
                batch_entries = batch_entries.split("///\n")
                # Remove possible empty last line
                batch_entries = [e for e in batch_entries if e.strip()]
                entries.extend(map(self.ENTRY_TYPE, batch_entries))
            start += batch_size

        return entries

    def _add_db(self, key):
        """
        Prefix the key with '%(DB)s:' string if not already prefixed.
        """
        if not key.startswith(self.DB + ":"):
            return self.DB + ":" + key
        else:
            return key


@entry.entry_decorate
class GenomeEntry(entry.DBEntry):
    """
    Entry for a KEGG Genome database.
    """

    FIELDS = [
        ("ENTRY", fields.DBEntryField),
        ("ORG_CODE", fields.DBSimpleField),
        ("NAME", fields.DBNameField),
        ("DEFINITION", fields.DBDefinitionField),
        ("ANNOTATION", fields.DBSimpleField),
        ("TAXONOMY", fields.DBTaxonomyField),
        ("DATA_SOURCE", fields.DBSimpleField),
        ("ORIGINAL_DB", fields.DBSimpleField),
        ("KEYWORDS", fields.DBSimpleField),
        ("DISEASE", fields.DBSimpleField),
        ("COMMENT", fields.DBSimpleField),
        ("CHROMOSOME", fields.DBFieldWithSubsections),
        ("PLASMID", fields.DBSimpleField),
        ("STATISTICS", fields.DBSimpleField),
        ("REFERENCE", fields.DBReference),
    ]

    MULTIPLE_FIELDS = ["REFERENCE"]

    def __init__(self, text):
        entry.DBEntry.__init__(self, text)

    @property
    def organism_code(self):
        """
        A three or four letter KEGG organism code (e.g. 'hsa', 'sce', ...)
        """
        return self.org_code

    @property
    def taxid(self):
        """
        Organism NCBI taxonomy id.
        """
        return self.TAXONOMY.taxid


class Genome(DBDataBase):
    """
    An interface to the A KEGG GENOME database.
    """

    DB = "genome"
    ENTRY_TYPE = GenomeEntry

    # For obiTaxonomy.common_taxids mapping
    TAXID_MAP = {
        "562": "511145",  # Escherichia coli K-12 MG1655
        "2104": "272634",  # Mycoplasma pneumoniae M129
        "4530": "39947",  # Oryza sativa ssp. japonica cultivar Nipponbare (Japanese rice)
        "4932": "559292",  # Saccharomyces cerevisiae S288C
        "4896": "284812",  # Schizosaccharomyces pombe 972h-
    }

    def __init__(self):
        DBDataBase.__init__(self)
        self._org_list = self.api.list_organisms()
        self._keys = [org.entry_id for org in self._org_list]

    def _key_to_gn_entry_id(self, key):
        res = self.find(key)
        if len(res) == 0:
            raise KeyError("Unknown key")
        elif len(res) > 1:
            raise ValueError("Not a unique key")
        else:
            return res[0]

    @classmethod
    def common_organisms(cls):
        return [
            'ath',
            'bta',
            'cel',
            'cre',
            'dre',
            'ddi',
            'dme',
            'eco',
            'hsa',
            'mmu',
            'mpn',
            'osa',
            'pfa',
            'rno',
            'sce',
            'spo',
            'zma',
            'xla',
        ]

    @classmethod
    def essential_organisms(cls):
        return ['ddi', 'dme', 'hsa', 'mmu', 'sce']

    def org_code_to_entry_key(self, code):
        """
        Map an organism code ('hsa', 'sce', ...) to the corresponding kegg
        identifier (T + 5 digit number).

        """
        for org in self._org_list:
            if org.org_code == code:
                return org.entry_id
        else:
            raise ValueError("Unknown organism code '%s'" % code)

    def search(self, string, relevance=False):
        """
        Search the genome database for string using ``bfind``.
        """
        if relevance:
            raise NotImplementedError("relevance is no longer supported")

        if string in self.TAXID_MAP:
            string = self.TAXID_MAP[string]

        res = self.api.find(self.DB, string).strip()
        if not res:
            return []

        res = res.splitlines()
        res = [r.split(",", 1)[0] for r in res]
        res = [r.split(None, 1)[1] for r in res]
        return res


@entry.entry_decorate
class GeneEntry(entry.DBEntry):
    FIELDS = [
        ("ENTRY", fields.DBEntryField),
        ("NAME", fields.DBNameField),
        ("DEFINITION", fields.DBDefinitionField),
        ("ORTHOLOGY", fields.DBSimpleField),
        ("ORGANISM", fields.DBSimpleField),
        ("PATHWAY", fields.DBPathway),
        ("MODULE", fields.DBSimpleField),
        ("BRITE", fields.DBSimpleField),
        ("DISEASE", fields.DBSimpleField),
        ("DRUG_TARGET", fields.DBSimpleField),
        ("CLASS", fields.DBSimpleField),
        ("MOTIF", fields.DBSimpleField),
        ("DBLINKS", fields.DBDBLinks),
        ("STRUCTURE", fields.DBSimpleField),
        ("POSITION", fields.DBSimpleField),
        ("AASEQ", fields.DBAASeq),
        ("NTSEQ", fields.DBNTSeq),
    ]

    def aliases(self):
        return (
            [self.entry_key]
            + (self.name.split(",") if self.name else [])
            + ([link[1][0] for link in self.dblinks.items()] if self.dblinks else [])
        )

    @property
    def alt_names(self):
        """
        For backwards compatibility.
        """
        return self.aliases()


class Genes(DBDataBase):
    """
    Interface to the KEGG Genes database.

    :param str org_code: KEGG organism code (e.g. 'hsa').

    """

    DB = None  # Needs to be set in __init__
    ENTRY_TYPE = GeneEntry

    def __init__(self, org_code):
        # TODO: Map to org code from kegg id (T + 5 digits)
        self.DB = org_code
        self.org_code = org_code
        DBDataBase.__init__(self)
        self._keys = self.api.get_genes_by_organism(org_code)

    def gene_aliases(self):
        aliases = {}
        for entry in self.itervalues():
            aliases.update(dict.fromkeys(entry.aliases(), self.org_code + ":" + entry.entry_key))

        return aliases


@entry.entry_decorate
class CompoundEntry(entry.DBEntry):
    FIELDS = [
        ("ENTRY", fields.DBEntryField),
        ("NAME", fields.DBNameField),
        ("FORMULA", fields.DBSimpleField),
        ("EXACT_MASS", fields.DBSimpleField),
        ("MOL_WEIGHT", fields.DBSimpleField),
        ("REMARK", fields.DBSimpleField),
        ("COMMENT", fields.DBSimpleField),
        ("REACTION", fields.DBSimpleField),
        ("PATHWAY", fields.DBPathway),
        ("ENZYME", fields.DBSimpleField),
        ("BRITE", fields.DBSimpleField),
        ("REFERENCE", fields.DBSimpleField),
        ("DBLINKS", fields.DBDBLinks),
        ("ATOM", fields.DBSimpleField),
        ("BOND", fields.DBSimpleField),
    ]


class Compound(DBDataBase):
    DB = "cpd"
    ENTRY_TYPE = CompoundEntry

    def __init__(self):
        DBDataBase.__init__(self)
        self._keys = [d.entry_id for d in self.api.list("cpd")]


@entry.entry_decorate
class ReactionEntry(entry.DBEntry):
    FIELDS = [
        ("ENTRY", fields.DBEntryField),
        ("NAME", fields.DBNameField),
        ("DEFINITION", fields.DBDefinitionField),
        ("EQUATION", fields.DBSimpleField),
        ("ENZYME", fields.DBSimpleField),
    ]


class Reaction(DBDataBase):
    DB = "rn"
    ENTRY_TYPE = ReactionEntry

    def __init__(self):
        DBDataBase.__init__(self)
        self._keys = [d.entry_id for d in self.api.list("rn")]


class Brite(DBDataBase):
    DB = "br"


class Disease(DBDataBase):
    DB = "ds"


class Drug(DBDataBase):
    DB = "dr"


@entry.entry_decorate
class EnzymeEntry(entry.DBEntry):
    FIELDS = [
        ("ENTRY", fields.DBEntryField),
        ("NAME", fields.DBNameField),
        ("CLASS", fields.DBSimpleField),
        ("SYSNAME", fields.DBSimpleField),
        ("REACTION", fields.DBSimpleField),
        ("ALL_REAC", fields.DBSimpleField),
        ("SUBSTRATE", fields.DBSimpleField),
        ("PRODUCT", fields.DBSimpleField),
        ("COMMENT", fields.DBSimpleField),
        ("REFERENCE", fields.DBReference),
        ("PATHWAY", fields.DBPathway),
        ("ORTHOLOGY", fields.DBSimpleField),
        ("GENES", fields.DBSimpleField),
        ("DBLINKS", fields.DBDBLinks),
    ]

    MULTIPLE_FIELDS = ["REFERENCE"]


class Enzyme(DBDataBase):
    DB = "ec"
    ENTRY_TYPE = EnzymeEntry

    def __init__(self):
        DBDataBase.__init__(self)
        self._keys = [d.entry_id for d in self.api.list("ec")]


@entry.entry_decorate
class OrthologyEntry(entry.DBEntry):
    FIELDS = [
        ("ENTRY", fields.DBEntryField),
        ("NAME", fields.DBNameField),
        ("CLASS", fields.DBSimpleField),
        ("DBLINKS", fields.DBDBLinks),
        ("GENES", fields.DBSimpleField),
    ]


class Orthology(DBDataBase):
    DB = "ko"
    ENTRY_TYPE = OrthologyEntry

    def __init__(self):
        DBDataBase.__init__(self)
        self._keys = [d.entry_id for d in self.api.list("ko")]


@entry.entry_decorate
class PathwayEntry(entry.DBEntry):
    FIELDS = [
        ("ENTRY", fields.DBEntryField),
        ("NAME", fields.DBNameField),
        ("DESCRIPTION", fields.DBSimpleField),
        ("CLASS", fields.DBSimpleField),
        ("PATHWAY_MAP", fields.DBPathwayMapField),
        ("MODULE", fields.DBSimpleField),
        ("DISEASE", fields.DBSimpleField),
        ("DRUG", fields.DBSimpleField),
        ("DBLINKS", fields.DBDBLinks),
        ("ORGANISM", fields.DBSimpleField),
        ("GENE", fields.DBGeneField),
        ("ENZYME", fields.DBEnzymeField),
        ("COMPOUND", fields.DBCompoundField),
        ("REFERENCE", fields.DBReference),
        ("REL_PATHWAY", fields.DBSimpleField),
        ("KO_PATHWAY", fields.DBSimpleField),
    ]

    MULTIPLE_FIELDS = ["REFERENCE"]

    @property
    def gene(self):
        if hasattr(self, "GENE"):
            genes = self.GENE._convert()
        else:
            return None

        org = self.organism
        org_prefix = ""
        if org:
            match = re.findall(r"\[GN:([a-z]+)\]", org)
            if match:
                org_prefix = match[0] + ":"
        genes = [org_prefix + g for g in genes]
        return genes


class Pathway(DBDataBase):
    """
    KEGG Pathway database

    :param str prefix:
        KEGG Organism code ('hsa', ...) or 'map', 'ko', 'ec' or 'rn'

    """

    DB = "path"
    ENTRY_TYPE = PathwayEntry

    def __init__(self, prefix="map"):
        DBDataBase.__init__(self)
        self.prefix = prefix
        valid = [d.org_code for d in self.api.list_organisms()] + ["map", "ko", "ec", "rn"]

        if prefix not in valid:
            raise ValueError("Invalid prefix %r" % prefix)

        self._keys = [d.entry_id for d in self.api.list("pathway/" + prefix)]



""" NCBI GeneInformation module """
import json
import sqlite3
import contextlib
from typing import Dict, List, Tuple, Optional

from Orange.data import Table, Domain, StringVariable
from Orange.data.util import get_unique_names_domain

from orangecontrib.bioinformatics.utils import serverfiles
from orangecontrib.bioinformatics.ncbi.taxonomy import species_name_to_taxid
from orangecontrib.bioinformatics.ncbi.gene.config import (
    DOMAIN,
    ENTREZ_ID,
    query,
    query_exact,
    gene_info_attributes,
)
from orangecontrib.bioinformatics.widgets.utils.data import TableAnnotation


class Gene:
    """Representation of gene summary."""

    __slots__ = gene_info_attributes + ('input_identifier',)

    def __init__(self, input_identifier: Optional[str] = None):
        """
        If we want to match gene to it's corresponding Entrez ID we must,
        upon class initialization, provide some `input identifier`. This way
        :class:`GeneMatcher` will know what to match it against in Gene Database.

        Parameters
        ----------
        input_identifier : str
            This can be any of the following: symbol, synonym, locus tag, other database id, ...
        """
        self.input_identifier = input_identifier

    def __getattr__(self, attribute):
        if attribute not in self.__slots__:
            return None

    def __repr__(self):
        return (
            f'<Gene symbol={self.symbol}, tax_id={self.tax_id}, gene_id={self.gene_id}>'
        )

    def load_attributes(
        self,
        values: Tuple[str, ...],
        attributes: Tuple[str, ...] = gene_info_attributes,
    ):
        for attr, val in zip(attributes, values):
            setattr(
                self,
                attr,
                json.loads(val) if attr in ('synonyms', 'db_refs', 'homologs') else val,
            )

    def homolog_gene(self, taxonomy_id: str) -> Optional[str]:
        """Returns gene homolog for given organism.

        Parameters
        ----------
        taxonomy_id: str
            Taxonomy id of target organism.

        Returns
        -------
        str
            Entrez ID (if available).
        """
        return self.homologs.get(taxonomy_id, None)


class GeneMatcher:
    """Gene name matching interface."""

    def __init__(self, tax_id: str, progress_callback=None, auto_start=True):
        """

        Parameters
        ----------
        tax_id:: str
            Taxonomy id of target organism.

        """
        self._tax_id: str = tax_id
        self._genes: List[Gene] = []
        self._progress_callback = progress_callback
        self._auto_start = auto_start
        self.gene_db_path = self._gene_db_path()

    @property
    def tax_id(self):
        return self._tax_id

    @tax_id.setter
    def tax_id(self, tax_id: str) -> None:
        self._tax_id = tax_id
        self.gene_db_path = self._gene_db_path()

    @property
    def genes(self) -> List[Gene]:
        return self._genes

    @genes.setter
    def genes(self, genes: List[str]) -> None:
        self._genes = [Gene(input_identifier=gene) for gene in genes]
        if self._auto_start:
            self._match()

    def get_known_genes(self) -> List[Gene]:
        """Return Genes with known Entrez ID

        Returns
        -------
        :class:`list` of :class:`Gene` instances
            Genes with unique match

        """
        return [gene for gene in self.genes if gene.gene_id]

    def to_data_table(self, selected_genes: Optional[List[str]] = None) -> Table:
        """Transform GeneMatcher results to Orange data table.

        Optionally we can provide a list of genes (Entrez Ids).
        The table on the output will be populated only with provided genes.

        Parameters
        ----------
        selected_genes: list
            List of Entrez Ids

        Returns
        -------
        Orange.data.Table
            Summary of Gene info in tabular format
        """
        data_x = []
        metas = [
            StringVariable('Input gene ID'),
            StringVariable(ENTREZ_ID),
            StringVariable('Symbol'),
            StringVariable('Synonyms'),
            StringVariable('Description'),
            StringVariable('Other IDs'),
            StringVariable('Type of gene'),
            StringVariable('Chromosome'),
            StringVariable('Map location'),
            StringVariable('Locus tag'),
            StringVariable('Symbol from nomenclature authority'),
            StringVariable('Full name from nomenclature authority'),
            StringVariable('Nomenclature status'),
            StringVariable('Other designations'),
            StringVariable('Species'),
            StringVariable('Taxonomy ID'),
        ]
        domain = Domain([], metas=metas)

        genes: List[Gene] = self.genes
        if selected_genes is not None:
            selected_genes_set = set(selected_genes)
            genes = [
                gene for gene in self.genes if str(gene.gene_id) in selected_genes_set
            ]

        for gene in genes:
            db_refs = (
                ', '.join(
                    '{}: {}'.format(key, val) for (key, val) in gene.db_refs.items()
                )
                if gene.db_refs
                else ''
            )
            synonyms = ', '.join(gene.synonyms) if gene.synonyms else ''

            line = [
                gene.input_identifier,
                gene.gene_id,
                gene.symbol,
                synonyms,
                gene.description,
                db_refs,
                gene.type_of_gene,
                gene.chromosome,
                gene.map_location,
                gene.locus_tag,
                gene.symbol_from_nomenclature_authority,
                gene.full_name_from_nomenclature_authority,
                gene.nomenclature_status,
                gene.other_designations,
                species_name_to_taxid(gene.species),
                gene.tax_id,
            ]

            data_x.append(line)

        table = Table.from_list(domain, data_x)
        table.name = 'Gene Matcher Results'
        table.attributes[TableAnnotation.tax_id] = self.tax_id
        table.attributes[TableAnnotation.gene_as_attr_name] = False
        table.attributes[TableAnnotation.gene_id_column] = ENTREZ_ID
        return table

    def match_table_column(
        self,
        data_table: Table,
        column_name: str,
        target_column: Optional[StringVariable] = None,
    ) -> Table:
        """Helper function for gene name matching with :class:`Orange.data.Table`.

        Give a column of genes, GeneMatcher will try to map genes to their
        corresponding Entrez Ids.


        Parameters
        ----------
        data_table: :class:`Orange.data.Table`
            Data table

        column_name: str
            Name of the column where gene symbols are located

        target_column: :class:`StringVariable`
            Column where we store Entrez Ids.
            Defaults to StringVariable(ncbi.gene.config.NCBI_ID)

        Returns
        -------
        :class:`Orange.data.Table`
            Data table with a column of Gene Ids
        """

        if column_name in data_table.domain:
            self.genes = data_table.get_column(column_name)

            if target_column is None:
                target_column = StringVariable(ENTREZ_ID)

            new_domain = Domain(
                data_table.domain.attributes,
                data_table.domain.class_vars,
                data_table.domain.metas + (target_column,),
            )

            new_data = data_table.transform(new_domain)
            with new_data.unlocked(new_data.metas):
                new_data[:, target_column] = [
                    [str(gene.gene_id) if gene.gene_id else '?'] for gene in self.genes
                ]

            return new_data

    def match_table_attributes(
        self, data_table, run=True, rename=False, source_name='Source ID'
    ) -> Table:
        """Helper function for gene name matching with :class:`Orange.data.Table`.

        Match table attributes and if a unique match is found create a
        new column attribute for Entrez Id. Attribute name is defined
        here: `orangecontrib.bioinformatics.ncbi.gene.config.NCBI_ID`


        Parameters
        ----------
        data_table: :class:`Orange.data.Table`
            Data table

        Returns
        -------

        :class:`Orange.data.Table`
            Data table column attributes are populated with Entrez Ids

        """
        # run gene matcher
        if run:
            self.genes = [var.name for var in data_table.domain.attributes]

        def helper(gene, attribute):
            if gene.gene_id:
                if rename:
                    attribute = attribute.renamed(gene.symbol)
                    attribute.attributes[source_name] = gene.input_identifier

                attribute.attributes[ENTREZ_ID] = gene.gene_id
            return attribute

        attributes = [
            helper(gene, attr)
            for gene, attr in zip(self.genes, data_table.domain.attributes)
        ]
        metas = data_table.domain.metas
        (attr_deduplicated, _, metas_deduplicated), renamed = get_unique_names_domain(
            [a.name for a in attributes], metas=[m.name for m in metas]
        )

        if len(renamed):
            attributes = [
                attr.renamed(new_name)
                for attr, new_name in zip(attributes, attr_deduplicated)
            ]
            metas = [
                meta.renamed(new_name)
                for meta, new_name in zip(metas, metas_deduplicated)
            ]

        domain = Domain(attributes, data_table.domain.class_vars, metas)
        return data_table.transform(domain)

    def match_genes(self):
        self._match()

    def _gene_db_path(self):
        return serverfiles.localpath_download(DOMAIN, f'{self.tax_id}.sqlite')

    def _match(self):
        synonyms, db_refs = 4, 5

        with contextlib.closing(sqlite3.connect(self.gene_db_path)) as con:
            with con as cursor:
                for gene in self.genes:
                    if self._progress_callback:
                        self._progress_callback()

                    search_param = gene.input_identifier.lower()

                    if search_param:
                        match_statement = (
                            '{gene_id symbol locus_tag symbol_from_nomenclature_authority}:^"'
                            + search_param
                            + '"'
                        )
                        match = cursor.execute(
                            query_exact, (match_statement,) + tuple([search_param] * 4)
                        ).fetchall()
                        # if unique match
                        if len(match) == 1:
                            gene.load_attributes(match[0])
                            continue

                        match = cursor.execute(
                            query, (f'synonyms:"{search_param}"',)
                        ).fetchall()
                        synonym_matched_rows = [
                            m
                            for m in match
                            if search_param
                            in (x.lower() for x in json.loads(m[synonyms]))
                        ]
                        # if unique match
                        if len(synonym_matched_rows) == 1:
                            gene.load_attributes(synonym_matched_rows[0])
                            continue

                        match = cursor.execute(
                            query, (f'db_refs:"{search_param}"',)
                        ).fetchall()
                        db_ref_matched_rows = [
                            m
                            for m in match
                            if search_param
                            in (x.lower() for x in json.loads(m[db_refs]).values())
                        ]
                        # if unique match
                        if len(db_ref_matched_rows) == 1:
                            gene.load_attributes(db_ref_matched_rows[0])
                            continue


class GeneInfo(dict):
    def __init__(self, tax_id: str):
        """Loads genes for given organism in a dict.

        Each instance of :class:`Gene` is mapped to corresponding Entrez ID

        Parameters
        ----------
        tax_id: str
            Taxonomy id of target organism.

        """
        super().__init__()
        self.tax_id: str = tax_id
        self.gene_db_path: str = self._gene_db_path()

        connection = sqlite3.connect(self.gene_db_path)
        cursor = connection.cursor()

        for gene_info in cursor.execute('SELECT * FROM gene_info').fetchall():
            gene = Gene()
            gene.load_attributes(gene_info)
            self[gene.gene_id] = gene

        cursor.close()
        connection.close()

    def _gene_db_path(self):
        return serverfiles.localpath_download(DOMAIN, f'{self.tax_id}.sqlite')


def load_gene_summary(tax_d: str, genes: List[Optional[str]]) -> List[Optional[Gene]]:
    gene_db_path = serverfiles.localpath_download(DOMAIN, f'{tax_d}.sqlite')

    # filter NoneTypes
    _genes = [g for g in genes if g]

    with contextlib.closing(sqlite3.connect(gene_db_path)) as con:
        with con as cur:
            gene_map: Dict[str, Gene] = {}
            for gene_info in cur.execute(
                f'SELECT * FROM gene_info WHERE gene_id in ({",".join(_genes)})'
            ).fetchall():
                gene = Gene()
                gene.load_attributes(gene_info)
                gene_map[gene.gene_id] = gene

            return [gene_map.get(gid, None) if gid else None for gid in genes]


if __name__ == "__main__":
    gm = GeneMatcher('9606')
    gm.genes = ['CD4', '614535', 'ENSG00000205426', "2'-PDE", 'HB-1Y']
    print(list(zip(gm.genes, [g.input_identifier for g in gm.genes])))
    _homologs = load_gene_summary(
        '10090', [g.homolog_gene(taxonomy_id='10090') for g in gm.genes]
    )
    print(_homologs)
